
__version__ = "0.7rc3"
__version_vcs_info__ = {
    'branch': 'release-0.7',
    'revid': 'd595255',
    'revno': 566}
__version_user_email__ = "skalkoto@grnet.gr"
__version_user_name__ = "Nikos Skalkotos"
